'''Generate random assay plates.'''

from argparse import Namespace
import csv
import json
from pathlib import Path
import random
import sys

from .params import AssayParams, load_params


MODEL = 'Weyland-Yutani 470'
PLATE_HEIGHT = 4
PLATE_WIDTH = 4


def plates(options: Namespace) -> None:
    '''Main driver for snailz plate creation.

    Each plate is represented by two files: one with the design and one with the results.
      Plate generation re-uses assay parameters: see `assays.py` for details.
      Plates are saved in the specified output directory with auto-generated names.

    -   options.assays: path to assay parameter files.
    -   options.designs: output directory for plate design files.
    -   options.readings: output directory for plate reading files.

    Args:
        options: see above.
    '''
    options.params = load_params(AssayParams, options.params)
    random.seed(options.params.seed)
    for filename, sample_id, kind in _join_assay_data(options):
        _make_plate(
            options.params,
            sample_id,
            kind,
            Path(options.designs, filename),
            Path(options.readings, filename),
        )


def _generate(params: AssayParams, sample_locs: list, func: callable) -> list:
    '''Make body of plate design or results.

    -   Each result is represented as a rectangular list of lists.
    -   The first row of each result is lettered started with 'A' like a spreadsheet.
    -   Each subsequent row starts with a row number from '1' like a spreadsheet.

    Args:
        params: assay parameters.
        sample_locs: location of sample in each row of plate
        func: function used to generate interior table value

    Returns:
        Rectangular list-of-lists.
    '''
    title_row = ['', *[chr(ord('A') + col) for col in range(PLATE_WIDTH)]]
    values = [
        [func(params, col == sample_locs[row]) for col in range(PLATE_WIDTH)]
        for row in range(PLATE_HEIGHT)
    ]
    labeled = [[str(i + 1), *r] for (i, r) in enumerate(values)]
    return [title_row, *labeled]


def _join_assay_data(options: Namespace) -> callable:
    '''Get experiment type and plate filename from data.

    Args:
        options: see above.

    Returns:
        Generator that produces tuples of plate filename, plate, and experiment.
    '''
    assays = json.load(open(options.assays, 'r'))
    experiments = {x['sample_id']: x['kind'] for x in assays['experiment']}
    plates = {p['filename']: p['sample_id'] for p in assays['plate']}
    return ((f, plates[f], experiments[plates[f]]) for f in plates)


def _make_head(kind: str, sample_id: int) -> list:
    '''Make head of plate.

    Args:
        kind: kind of experiment ('calibration' or other).
        sample_id: which sample this experiment is for.

    Returns:
        List-of-lists representation of head of plate.
    '''
    return [
        [MODEL, kind, sample_id],
        [],
    ]


def _make_placement(kind: str) -> tuple:
    '''Generate random arrangement of sample locations in plate.

    Args:
        kind: kind of experiment ('calibration' or other).

    Returns:
        Tuple of sample placements and column orders.
    '''
    placement = [[False for col in range(PLATE_WIDTH)] for row in range(PLATE_HEIGHT)]
    if kind == 'calibration':
        return placement, []
    columns = list(c for c in range(PLATE_WIDTH))
    random.shuffle(columns)
    columns = columns[:PLATE_HEIGHT]
    for r, row in enumerate(placement):
        row[columns[r]] = True
    return placement, columns


def _make_plate(params: AssayParams, sample_id: str, kind: str, design_file: str, readings_file: str) -> None:
    '''Generate an entire experimental plate.

    1.  Build overall placement grid and sample locations.
    2.  Make head and body of design file and save it.
    3.  Make random readings for plate cells and save.
    4.  Save plate and design in respective output directories.

    Args:
        params: assay parameters.
        sample_id: which sample this plate is for.
        kind: 'calibration' or something else.
        design_file: where to write design.
        readings_file: where to write readings.
    '''
    placement, sample_locs = _make_placement(kind)

    design = [*_make_head('design', sample_id), *_generate(params, sample_locs, _make_treatment)]
    _save(design_file, _normalize_csv(design))

    readings = [*_make_head('readings', sample_id), *_generate(params, sample_locs, _make_reading)]
    _save(readings_file, _normalize_csv(readings))


def _make_reading(params: AssayParams, treated: bool) -> float:
    '''Generate a single plate reading.

    The reading is a random value whose mean depends on whether it is treated or a control.

    Args:
        params: assay parameters.
        treated: is this location treated (versus a control)?

    Returns:
        Randomly-generated reading value.
    '''
    mean = params.treated_val if treated else params.control_val
    value = max(0.0, random.gauss(mean, params.stdev))
    return f'{value:.02f}'


def _make_treatment(params: AssayParams, treated: bool) -> str:
    '''Select a single plate treatment.

    Args:
        params: assay parameters.
        treated: is this location treated (versus a control)?

    Returns:
        Treatment if this is a treated cell or a randomly-selected control if it is not.
    '''
    result = params.treatment if treated else random.choice(params.controls)
    return result


def _normalize_csv(rows: list) -> list:
    '''Make sure all rows in list-of-lists table are the same length.

    Short rows are extended with empty string values to match the length of the longest row.

    Args:
        rows: input list of lists.

    Returns:
        Adjusted list of lists.
    '''
    required = max(len(r) for r in rows)
    for row in rows:
        row.extend([''] * (required - len(row)))
    return rows


def _save(filename: str|None, rows: list) -> None:
    '''Save results to file or show on standard output.

    Args:
        filename: output file (or None to write to standard output)
        rows: data to write.
    '''
    if not filename:
        csv.writer(sys.stdout).writerows(rows)
    else:
        csv.writer(open(filename, 'w'), lineterminator='\n').writerows(rows)



'''Parameter dataclasses and utilities.'''

from argparse import Namespace
from dataclasses import dataclass, field
from datetime import date, datetime, timezone
from importlib.resources import files
import json
from pathlib import Path
from typing import List
import pytz

# Use ISO date format.
DATE_FORMAT = '%Y-%m-%d'

# Start and end dates for experiemnts.
DEFAULT_START_DATE = datetime.strptime('2023-11-01', DATE_FORMAT).replace(tzinfo=pytz.UTC)
DEFAULT_END_DATE = datetime.strptime('2023-11-10', DATE_FORMAT).replace(tzinfo=pytz.UTC)

# Parameter files to include in package.
PARAMETER_FILES = (
    'params/assays.json',
    'params/genomes.json',
    'params/grids.json',
    'params/samples.json',
    'params/sites.csv',
    'params/staff.json',
    'params/surveys.csv',
)


@dataclass
class AssayParams:
    '''Parameters for assay data generation.'''

    assay_staff: list
    assay_types: list
    assay_duration: list
    assay_plates: list
    control_val: float = 5.0
    controls: List[str] = field(default_factory=list)
    enddate: str = None
    experiments: int = 1
    filename_length: int = 8
    fraction: float = None
    invalid: float = 0.1
    seed: int = None
    startdate: str = None
    stdev: float = 3.0
    treated_val: float = 8.0
    treatment: str = None

    def __post_init__(self):
        '''Fill in missing dates and convert to standard format.'''
        if self.startdate is None:
            self.startdate = DEFAULT_START_DATE
        elif isinstance(self.startdate, str):
            self.startdate = datetime.strptime(self.startdate, DATE_FORMAT).replace(tzinfo=pytz.UTC)
        elif not self.startdate.tzinfo:
            self.startdate = self.startdate.replace(tzinfo=pytz.UTC)

        if self.enddate is None:
            self.enddate = DEFAULT_END_DATE
        elif isinstance(self.enddate, str):
            self.enddate = datetime.strptime(self.enddate, DATE_FORMAT).replace(tzinfo=pytz.UTC)
        elif not self.enddate.tzinfo:
            self.enddate = self.enddate.replace(tzinfo=pytz.UTC)


@dataclass
class GenomeParams:
    '''Gene sequence parameters.'''
    snp_probs: list
    length: int
    num_genomes: int
    num_snp: int
    prob_other: float
    seed: int = None


@dataclass
class GridParams:
    '''Invasion percolation parameters.'''
    depth: int
    height: int
    seed: int
    width: int


@dataclass
class SampleParams:
    '''Sampled snail parameters.'''
    min_snail_size: float = None
    max_snail_size: float = None
    mutant: float = None
    normal: float = None
    seed: int = None


@dataclass
class StaffParams:
    '''Staff parameters.'''
    locale: str = 'et_EE'
    num: int = None
    seed: int = None


def export_params(options: Namespace) -> None:
    '''Export parameter files.

    This function is used to write default parameter files when the package is installed.

    -   options.outdir: directory to write to.
    '''
    outdir = Path(options.outdir)
    outdir.mkdir(parents=True, exist_ok=True)
    root = files(__name__.split('.')[0])
    for filename in PARAMETER_FILES:
        src = root.joinpath(filename)
        dst = outdir.joinpath(Path(filename).name)
        dst.write_bytes(src.read_bytes())


def load_params(cls: type, filename: str) -> object:
    '''Load parameters from file and return as object.

    Args:
        cls: class to instantiate.
        filename: file containing JSON representation of parameters.

    Returns:
        Populated instance of the given class.
    '''
    return cls(**json.loads(Path(filename).read_text()))



'''Generate genomes for snailz with random mutations.'''

from argparse import Namespace
from dataclasses import dataclass, asdict
import json
from pathlib import Path
import random

from .params import GenomeParams, load_params


# Bases.
DNA = 'ACGT'


@dataclass
class GenePool:
    '''Keep track of generated genomes.'''

    length: int
    reference: str
    individuals: list[str]
    locations: list[int]
    susceptible_loc: int = 0
    susceptible_base: str = ''


def genomes(options: Namespace) -> None:
    '''Main driver for genome generation.

    Each genome is a string of ACGT bases of the same length.
      One location is randomly chosen as "significant",
      and a specific mutation there predisposes the snail to size changes.
      Other mutations are added randomly at other locations.

    - options.params: parameter file.
    - options.outfile: output file.

    The result is saved as JSON with the following entries:

    - length: fixed length of all genomes.
    - reference: the unmutated reference genome.
    - individuals: a list of individual genomes with mutations.
    - locations: a list of locations where mutations may occur.
    - susceptible_loc: one of those locations where the significant mutation may occur.
    - susceptible_base: the mutated base at that location that indicates susceptibility.
    '''
    assert options.params != options.outfile, 'Cannot use same filename for options and parameters'
    options.params = load_params(GenomeParams, options.params)
    random.seed(options.params.seed)
    data = _random_genomes(options.params)
    _add_susceptibility(data)
    _save(options.outfile, data)


def _add_susceptibility(data: GenePool) -> None:
    '''Add indication of genetic susceptibility.

    Args:
        data: a GenePool instance being populated.
    '''
    if not data.locations:
        return
    loc = _choose_one(data.locations)
    choices = {ind[loc] for ind in data.individuals} - {data.reference[loc]}
    data.susceptible_loc = loc
    data.susceptible_base = _choose_one(list(sorted(choices)))


def _random_bases(length: int) -> str:
    '''Generate a random sequence of bases of the specified length.

    Args:
        length: desired genome length.

    Returns:
        Random sequence of bases of required length.
    '''
    assert 0 < length
    return ''.join(random.choices(DNA, k=length))


def _random_genomes(params: GenomeParams) -> GenePool:
    '''Generate a set of genomes with specified number of point mutations.

    Args:
        params: genome generation parameters.

    Returns:
        A GenePool object suitable for serialization.
    '''
    assert 0 <= params.num_snp <= params.length

    # Reference genomes and specific genomes to modify.
    reference = _random_bases(params.length)
    individuals = [reference] * params.num_genomes

    # Locations for SNPs.
    locations = random.sample(list(range(params.length)), params.num_snp)

    # Introduce significant mutations.
    for loc in locations:
        candidates = _other_bases(reference, loc)
        bases = [reference[loc]] + random.sample(candidates, k=len(candidates))
        individuals = [_mutate_snps(params, reference, ind, loc, bases) for ind in individuals]

    # Introduce other random mutations.
    other_locations = list(set(range(params.length)) - set(locations))
    individuals = [
        _mutate_other(ind, params.prob_other, other_locations) for ind in individuals
    ]

    # Return structure.
    individuals.sort()
    locations.sort()
    return GenePool(
        length=params.length, reference=reference, individuals=individuals, locations=locations
    )


def _save(outfile: str, data: GenePool) -> None:
    '''Save or show generated data.

    Args:
        outfile: output filename.
        data: to be saved.
    '''
    as_text = json.dumps(asdict(data), indent=4)
    if outfile:
        Path(outfile).write_text(as_text)
    else:
        print(as_text)


def _mutate_snps(params: GenomeParams, reference: str, genome: str, loc: int, bases: str) -> str:
    '''Introduce single nucleotide polymorphisms at the specified location.

    Args:
        params: genome generation parameters.
        reference: reference genome.
        genome: genome to mutate.
        loc: where to introduce mutation.
        bases: alternative bases.

    Returns:
        Mutated genome.
    '''
    choice = _choose_one(bases, params.snp_probs)
    return genome[:loc] + choice + genome[loc + 1 :]


def _mutate_other(genome: str, prob: float, locations: list) -> str:
    '''Introduce other mutations at specified locations.

    Args:
        genome: to be mutated.
        prob: probability of mutation.
        locations: where mutation might occur

    Returns:
        Possibly-mutated genome.
    '''
    if random.random() > prob:
        return genome
    loc = random.sample(locations, k=1)[0]
    base = random.choice(_other_bases(genome, loc))
    genome = genome[:loc] + base + genome[loc + 1 :]
    return genome


def _choose_one(values: list, weights: list|None = None) -> object:
    '''Convenience wrapper to choose a single items with weighted probabilities.

    Args:
        values: what to choose from.
        weights: optional list of weights.

    Returns:
        One value chosen at random from those given.
    '''
    return random.choices(values, weights=weights, k=1)[0]


def _other_bases(seq: str, loc: int) -> list:
    '''Create a list of bases minus the one in the sequence at that location.

    Returns a list instead of a set because the result is used in random.choices(),
      which requires an indexable sequence. Result is sorted for reproducibility.

    Args:
        seq: base sequence.
        loc: location of base to _not_ choose.

    Returns:
        List of other bases.
    '''
    return list(sorted(set(DNA) - {seq[loc]}))



'''Create map of snailz survey points based on generated data.'''

from argparse import Namespace
import plotly.graph_objects as go
import polars as pl


def surveymap(options: Namespace) -> None:
    '''Main driver for snailz map creation.

    -   options.samples: path to samples CSV file.
    -   options.outfile: optional path to saved output file.

    Args:
        options: see list above.

    Returns:
        Either writes the generated map to a file or displays it in the browser.
    '''
    samples = pl.read_csv(options.samples)['survey_id', 'lon', 'lat']
    fig = go.Figure(go.Scattermapbox(
        lon=samples['lon'],
        lat=samples['lat'],
        marker=go.scattermapbox.Marker(color=samples['survey_id']),
    ))
    fig.update_layout(
        mapbox={
            'style': 'open-street-map',
            'center': {'lon': -124.2, 'lat': 48.85},
            'zoom': 11,
        },
        margin={'l': 0, 'r': 0, 'b': 0, 't': 0},
    )
    if options.outfile:
        fig.write_image(options.outfile)
    else:
        fig.show()



'''Interface for command-line script.'''

import argparse
import importlib.metadata
from pathlib import Path

from .assays import assays
from .db import db
from .genomes import genomes
from .grid import grid
from .mangle import mangle
from .plates import plates
from .samples import samples
from .staff import staff
from .surveymap import surveymap

from .params import export_params


def main():
    '''Main driver.'''
    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='store_true', help='show version')
    subparsers = parser.add_subparsers()
    for sub in (
            _assays_parser,
            _db_parser,
            _everything_parser,
            _genomes_parser,
            _grid_parser,
            _mangle_parser,
            _params_parser,
            _plates_parser,
            _samples_parser,
            _staff_parser,
            _surveymap_parser,
    ):
        sub(subparsers)
    options = parser.parse_args()
    if options.version:
        print(importlib.metadata.version('snailz'))
    else:
        options.func(options)


def everything(options: argparse.Namespace) -> None:
    '''Build everything using default values.

    Args:
        options: controlling options.

    -   paramsdir: input parameters directory.
    -   datadir: output data directory.
    '''
    # Common values
    assays_data = Path(options.datadir, 'assays.json')
    assays_params = Path(options.paramsdir, 'assays.json')
    db_file = Path(options.datadir, 'lab.db')
    designs_data_dir = Path(options.datadir, 'designs')
    genomes_data = Path(options.datadir, 'genomes.json')
    genomes_params = Path(options.paramsdir, 'genomes.json')
    grids_data_dir = Path(options.datadir, 'grids')
    grids_params = Path(options.paramsdir, 'grids.json')
    mangled_data_dir = Path(options.datadir, 'mangled')
    readings_data_dir = Path(options.datadir, 'readings')
    samples_data = Path(options.datadir, 'samples.csv')
    samples_params = Path(options.paramsdir, 'samples.json')
    sites_params = Path(options.paramsdir, 'sites.csv')
    staff_data = Path(options.datadir, 'staff.csv')
    staff_params = Path(options.paramsdir, 'staff.json')
    surveys_params = Path(options.paramsdir, 'surveys.csv')
    surveymap_file = Path(options.datadir, 'survey.png')

    # Ensure root data directory exists
    _verbose(options, 'data directory')
    Path(options.datadir).mkdir(exist_ok=True)

    # Grids
    _verbose(options, 'grids')
    grids_data_dir.mkdir(exist_ok=True)
    grid(_make_options(
        outdir=grids_data_dir,
        params=grids_params,
        sites=sites_params,
    ))

    # Genomes
    _verbose(options, 'genomes')
    genomes(_make_options(
        params=genomes_params,
        outfile=genomes_data,
    ))

    # Samples
    _verbose(options, 'samples')
    samples(_make_options(
        genomes=genomes_data,
        grids=grids_data_dir,
        outfile=samples_data,
        params=samples_params,
        sites=sites_params,
        surveys=surveys_params,
    ))

    # Staff
    _verbose(options, 'staff')
    staff(_make_options(
        params=staff_params,
        outfile=staff_data,
    ))

    # Assays
    _verbose(options, 'assays')
    assays(_make_options(
        genomes=genomes_data,
        outfile=assays_data,
        params=assays_params,
        samples=samples_data,
        staff=staff_data,
    ))

    # Plates
    _verbose(options, 'plates')
    designs_data_dir.mkdir(exist_ok=True)
    readings_data_dir.mkdir(exist_ok=True)
    plates(_make_options(
        designs=designs_data_dir,
        readings=readings_data_dir,
        assays=assays_data,
        params=assays_params,
    ))

    # Database
    _verbose(options, 'db')
    db(_make_options(
        dbfile=db_file,
        assays=assays_data,
        samples=samples_data,
        sites=sites_params,
        staff=staff_data,
        surveys=surveys_params,
    ))

    # Mangled plate files
    _verbose(options, 'mangled')
    mangled_data_dir.mkdir(exist_ok=True)
    mangle(_make_options(
        dbfile=db_file,
        tidy=readings_data_dir,
        outdir=mangled_data_dir,
    ))

    # Survey map
    if options.withmap:
        _verbose(options, 'survey map')
        surveymap(_make_options(
            outfile=surveymap_file,
            samples=samples_data,
        ))


def _assays_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for assays.'''
    parser = subparsers.add_parser('assays', help='construct assays')
    parser.add_argument('--genomes', type=str, required=True, help='genome file')
    parser.add_argument('--outfile', type=str, default=None, help='output file')
    parser.add_argument('--params', type=str, required=True, help='parameter file')
    parser.add_argument('--samples', type=str, required=True, help='samples file')
    parser.add_argument('--staff', type=str, required=True, help='staff file')
    parser.set_defaults(func=assays)


def _db_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for database creation.'''
    parser = subparsers.add_parser('db', help='construct database')
    parser.add_argument('--assays', type=str, required=True, help='assay data file')
    parser.add_argument('--dbfile', type=str, required=True, help='output database file')
    parser.add_argument('--samples', type=str, required=True, help='samples data file')
    parser.add_argument('--sites', type=str, required=True, help='sites parameter file')
    parser.add_argument('--surveys', type=str, required=True, help='surveys parameter file')
    parser.set_defaults(func=db)


def _everything_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for creating everything.'''
    parser = subparsers.add_parser('everything', help='build everything with defaults')
    parser.add_argument('--datadir', type=str, required=True, help='output data directory')
    parser.add_argument('--paramsdir', type=str, required=True, help='input parameters directory')
    parser.add_argument('--withmap', action='store_true', help='also generate map')
    parser.add_argument('--verbose', action='store_true', help='report progress')
    parser.set_defaults(func=everything)


def _genomes_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for creating genomes.'''
    parser = subparsers.add_parser('genomes', help='construct genomes')
    parser.add_argument('--outfile', type=str, default=None, help='output file')
    parser.add_argument('--params', type=str, required=True, help='parameter file')
    parser.set_defaults(func=genomes)


def _grid_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for creating grids.'''
    parser = subparsers.add_parser('grid', help='construct survey grid')
    parser.add_argument('--outdir', type=str, required=True, help='output directory')
    parser.add_argument('--params', type=str, required=True, help='grid parameter file')
    parser.add_argument('--sites', type=str, required=True, help='site parameter file')
    parser.set_defaults(func=grid)


def _mangle_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for mangling plate readings files.'''
    parser = subparsers.add_parser('mangle', help='mangle readings files')
    parser.add_argument('--dbfile', type=str, required=True, help='database file')
    parser.add_argument('--outdir', type=str, required=True, help='output directory')
    parser.add_argument('--tidy', type=str, required=True, help='input directory')
    parser.set_defaults(func=mangle)


def _params_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for copying parameters from installed package.'''
    parser = subparsers.add_parser('params', help='export parameter files')
    parser.add_argument('--outdir', type=str, required=True, help='output directory')
    parser.set_defaults(func=export_params)


def _plates_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for generating plates.'''
    parser = subparsers.add_parser('plates', help='construct plates')
    parser.add_argument('--assays', type=str, required=True, help='assays file')
    parser.add_argument('--designs', type=str, required=True, help='designs directory')
    parser.add_argument('--params', type=str, required=True, help='parameter file')
    parser.add_argument('--readings', type=str, required=True, help='readings directory')
    parser.set_defaults(func=plates)


def _samples_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for generating samples.'''
    parser = subparsers.add_parser('samples', help='construct samples')
    parser.add_argument('--genomes', type=str, required=True, help='genome file')
    parser.add_argument('--grids', type=str, required=True, help='grids directory')
    parser.add_argument('--outfile', type=str, help='output file')
    parser.add_argument('--params', type=str, required=True, help='parameter file')
    parser.add_argument('--sites', type=str, required=True, help='sites parameter file')
    parser.add_argument('--surveys', type=str, required=True, help='surveys parameter file')
    parser.set_defaults(func=samples)


def _staff_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for generating staff.'''
    parser = subparsers.add_parser('staff', help='construct staff names and IDs')
    parser.add_argument('--params', type=str, required=True, help='parameter file')
    parser.add_argument('--outfile', type=str, required=True, help='output file name')
    parser.set_defaults(func=staff)


def _surveymap_parser(subparsers: argparse._SubParsersAction) -> None:
    '''Add sub-parser for generating map of survey sites.'''
    parser = subparsers.add_parser('surveymap', help='construct survey locations')
    parser.add_argument('--outfile', type=str, required=True, help='output file name')
    parser.add_argument('--samples', type=str, required=True, help='samples data file')
    parser.set_defaults(func=surveymap)


def _make_options(**kwargs: dict) -> argparse.Namespace:
    '''Build an argparse options object.

    Args:
        kwargs: fields and values in options object.

    Returns:
        Options object.
    '''
    options = argparse.Namespace()
    for key, value in kwargs.items():
        setattr(options, key, value)
    return options


def _verbose(options: argparse.Namespace, msg: str) -> None:
    '''Possibly report progress.

    Args:
        options: controlling options (must have `verbose` field).
        msg: what to (possibly) print.
    '''
    if options.verbose:
        print(f'...{msg}')



'''Generate snailz experimental staff by creating random first and last names.'''

from argparse import Namespace
import faker
from pathlib import Path
import polars as pl
import random
import sys

from .params import StaffParams, load_params


def staff(options: Namespace) -> None:
    '''Main driver for snailz experimental staff creation.

    -   options.params: path to parameter file (see params.StaffParams for fields).
    -   options.outfile: optional path to saved output file.

    Generated data is written as CSV to the specified output file.

    Args:
        options: see above.
    '''
    options.params = load_params(StaffParams, options.params)
    random.seed(options.params.seed)
    faker.Faker.seed(options.params.seed)
    fake = faker.Faker(options.params.locale)
    people = _make_people(options.params, fake)
    _save(options, people)


def _make_people(params: StaffParams, fake: faker.Faker) -> pl.DataFrame:
    '''Create people.

    Args:
        params: staff generation parameters.
        fake: fake name generator.

    Returns:
        Dataframe containing staff ID, personal name, and family name.
    '''
    people = [(i+1, fake.first_name(), fake.last_name()) for i in range(params.num)]
    return pl.DataFrame(people, schema=('staff_id', 'personal', 'family'), orient='row')


def _save(options: Namespace, people: pl.DataFrame) -> None:
    '''Save results to file or show on standard output.

    Args:
        options: controlling options.
        people: dataframe of staff ID, personal name, and family name.
    '''
    if options.outfile:
        people.write_csv(Path(options.outfile))
    else:
        people.write_csv(sys.stdout)



'''Generate snailz sample grids using invasion percolation.

To create a grid:

1.  Generate an MxN grid of random numbers. (Our are always square and odd-sized.)
2.  Mark the center cell as "filled" by negating its value.
3.  On each iteration:
    1.  Find the lowest-valued cell adjacent to the filled region.
    2.  Fill that in by negating its value. (If several cells tie for lowest value, pick one at random.)
4.  Stop when the filled region hits the edge of the grid.

Instead of repeatedly searching for cells adjacent to the filled region,
  the grid keeps a value-to-coordinates dictionary.
  When a cell is filled in, neighbors not already recorded are added.

The grid is saved as comma-separated values.
  0 shows unfilled cells; filled cells contain their original (positive) value.
'''

from argparse import Namespace
from pathlib import Path
import polars as pl
import random

from .params import GridParams, load_params


class Grid:
    '''Represent a 2D grid that supports lazy filling.'''

    def __init__(self, params: GridParams) -> None:
        '''Record shared state.

        Args:
            params: grid creation parameters.
        '''
        self._width = params.width
        self._height = params.height
        self._depth = params.depth
        self._grid = []
        for x in range(self._width):
            col = []
            for y in range(self._height):
                col.append(random.randint(1, self._depth))
            self._grid.append(col)
        self._candidates = {}

    def __getitem__(self, key: list|tuple) -> int:
        '''Get value at location.

        Args:
            key: (x, y) coordinates.

        Returns:
            Value at that location.
        '''
        x, y = key
        return self._grid[x][y]

    def __setitem__(self, key: list|tuple, value: int) -> None:
        '''Set value at location.

        Args:
            key: (x, y) coordinates.
            value: new value.
        '''
        x, y = key
        self._grid[x][y] = value

    def __str__(self) -> str:
        '''Convert to printable string.'''
        rows = []
        for y in range(self.height - 1, -1, -1):
            rows.append(''.join('x' if self[x, y] == 0 else '.' for x in range(self.width)))
        return '\n'.join(rows)

    @property
    def depth(self) -> int:
        '''Get depth of grid.'''
        return self._depth

    @property
    def height(self) -> int:
        '''Get height of grid.'''
        return self._height

    @property
    def width(self) -> int:
        '''Get width of grid.'''
        return self._width

    def fill(self) -> None:
        '''Fill grid one cell at a time.'''
        x, y = self.width // 2, self.height // 2
        self[x, y] = - self[x, y]
        self.add_candidates(x, y)
        while True:
            x, y = self.choose_cell()
            self[x, y] = - self[x, y]
            if self.on_border(x, y):
                break

    def add_candidates(self, x: int, y: int) -> None:
        '''Add candidates next to a newly-filled cell.

        Args:
            x: X coordinate of newly-filled cell.
            y: Y coordinate of newly-filled cell.

        '''
        for ix in (x - 1, x + 1):
            self.add_one_candidate(ix, y)
        for iy in (y - 1, y + 1):
            self.add_one_candidate(x, iy)

    def add_one_candidate(self, x: int, y: int) -> None:
        '''Add a point to the set of candidates.

        Args:
            x: X coordinate of potential candidate.
            y: Y coordinate of potential candidate.
        '''
        if (x < 0) or (x >= self.width) or (y < 0) or (y >= self.height):
            return
        if self[x, y] < 0:
            return

        value = self[x, y]
        if value not in self._candidates:
            self._candidates[value] = set()
        self._candidates[value].add((x, y))

    def adjacent(self, x: int, y: int) -> bool:
        '''Is (x, y) adjacent to a filled cell?

        Args:
            x: X coordinate of cell to check.
            y: Y coordinate of cell to check.

        Returns:
            Whether or not this cell is adjacent to a filled cell.
        '''
        x_1, y_1 = x + 1, y + 1
        if (x > 0) and (self[x - 1, y] < 0):
            return True
        if (x_1 < self.width) and (self[x_1, y] < 0):
            return True
        if (y > 0) and (self[x, y - 1] < 0):
            return True
        if (y_1 < self.height) and (self[x, y_1] < 0):
            return True
        return False

    def choose_cell(self) -> tuple:
        '''Choose the next cell to fill.

        Returns:
            (x, y) coordinates of next cell to fill.
        '''
        min_key = min(self._candidates.keys())
        available = list(sorted(self._candidates[min_key]))
        i = random.randrange(len(available))
        choice = available[i]
        del available[i]
        if not available:
            del self._candidates[min_key]
        else:
            self._candidates[min_key] = set(available)
        self.add_candidates(*choice)
        return choice

    def on_border(self, x: int, y: int) -> bool:
        '''Is this cell on the border of the grid?

        Args:
            x: X coordinate of cell to check.
            y: Y coordinate of cell to check.

        Returns:
            Whether or not this cell is on the border of the grid.
        '''
        if (x == 0) or (x == self.width - 1):
            return True
        if (y == 0) or (y == self.height - 1):
            return True
        return False


def grid(options: Namespace) -> None:
    '''Main driver for grid generation.

    -   options.outdir: path to output directory.
    -   options.params: path to grid parameter file.
    -   options.sites: path to sites parameter file.

    Args:
        options: see above.
    '''
    options.params = load_params(GridParams, options.params)
    options.sites = pl.read_csv(options.sites)
    random.seed(options.params.seed)
    for row in options.sites.iter_rows(named=True):
        grid = Grid(options.params)
        grid.fill()
        _save(options.outdir, row['site_id'], grid)


def _save(outdir: str, site_id: str, grid: Grid) -> None:
    '''Save grid as CSV.

    Args:
        outdir: output directory.
        site_id: site identifier (used to construct output filename).
        grid: what to save.
    '''
    Path(outdir).mkdir(parents=True, exist_ok=True)
    with open(Path(outdir, f'{site_id}.csv'), 'w') as writer:
        for y in range(grid.height - 1, -1, -1):
            values = (- grid[x, y] if grid[x, y] < 0 else 0 for x in range(grid.width))
            print(','.join((str(v) for v in values)), file=writer)



'''Generate snailz assays.'''

from argparse import Namespace
from datetime import date, datetime, time, timedelta, timezone
import json
from pathlib import Path
import polars as pl
import random
import string
import pytz

from .params import AssayParams, load_params


class DateTimeEncoder(json.JSONEncoder):
    '''Encode date and datetime objects as JSON.'''

    def default(self, obj: date|datetime) -> str:
        '''Encode date or datetime.

        Args:
            obj: what to encode.

        Returns:
            String representation.
        '''
        if isinstance(obj, (date, datetime)):
            return obj.isoformat()


def assays(options: Namespace) -> None:
    '''Main driver for assay generation.

    1.  Load previously-generated samples and staff.
    2.  Generate experiments for some samples, each performed by one or more staff.
    3.  Save.

    Args:
        options: options object.
    '''
    assert options.params != options.outfile, 'Cannot use same filename for options and parameters'
    options.params = load_params(AssayParams, options.params)
    random.seed(options.params.seed)

    mutant_status = _reload_mutant_status(options)
    staff_ids = _reload_staff(options)
    result = _make_experiments(options.params, mutant_status, staff_ids)
    _save(options.outfile, result)


def _reload_mutant_status(options: Namespace) -> list:
    '''Re-create sample genomic information.

    Args:
        options: controlling options.

    Returns:
        List of Booleans showing which samples are for mutant snails.
    '''
    genomes = json.loads(Path(options.genomes).read_text())
    susc_loc = genomes['susceptible_loc']
    susc_base = genomes['susceptible_base']
    samples = pl.read_csv(options.samples)
    return [g[susc_loc] == susc_base for g in samples['sequence']]


def _reload_staff(options: Namespace) -> pl.DataFrame:
    '''Re-load staff information.

    Args:
        options: controlling options.

    Returns:
        Dataframe with staff information.
    '''
    return pl.read_csv(options.staff)['staff_id'].to_list()


def _make_experiments(params: AssayParams, mutant_status: list, staff_ids: list) -> dict:
    '''Create experiments and their data.

    Args:
        params: assay generation parameters.
        mutant_status: list showing which samples are mutants.
        staff_ids: list of staff IDs.

    Returns:
        Dictionary holding data to serialize.
    '''

    # Setup.
    kinds = list(params.assay_types)
    experiments = []
    performed = []
    plates = []

    # Which samples have experiments been done on?
    num_samples = len(mutant_status)
    keepers = set(random.sample(list(range(num_samples)), k=int(params.fraction * num_samples)))

    filename_gen = _make_random_filename_generator(params)
    for i, flag in enumerate(mutant_status):
        # Skip samples that aren't experimented on.
        if i not in keepers:
            continue

        # Create basic facts about experiment.
        sample_id = i + 1
        kind = random.choice(kinds)
        started, ended = _random_experiment_duration(params, kind)
        experiments.append(
            {'sample_id': sample_id, 'kind': kind, 'start': _round_date(started), 'end': _round_date(ended)}
        )

        # Keep track of staff involved in experiment.
        num_staff = random.randint(*params.assay_staff)
        performed.extend(
            [{'staff_id': s, 'sample_id': sample_id} for s in random.sample(staff_ids, num_staff)]
        )

        # Only generate plates for experiments that have finished.
        if ended is not None:
            plates.extend(
                _random_plates(params, kind, sample_id, len(plates), started, filename_gen)
            )

    # Invalidate some plates.
    invalidated = _invalidate_plates(params, staff_ids, plates)

    # Return structure with results.
    return {
        'experiment': experiments,
        'performed': performed,
        'plate': plates,
        'invalidated': invalidated
    }


def _invalidate_plates(params: AssayParams, staff_ids: int, plates: list) -> list:
    '''Invalidate a random set of plates.

    Args:
        params: assay generation parameters.
        staff_ids: list of staff IDs.
        plates: list of generated plates.

    Returns:
        List of dictionaries describing invalidated plates.
    '''
    selected = [
        (i, p['date']) for (i, p) in enumerate(plates) if random.random() < params.invalid
    ]
    return [
        {
            'plate_id': plate_id,
            'staff_id': random.choice(staff_ids),
            'date': _random_date_interval(exp_date, params.enddate),
        }
        for (plate_id, exp_date) in selected
    ]


def _make_random_filename_generator(params: AssayParams) -> callable:
    '''Create a random filename generator.

    Args:
        params: assay generation parameters.

    Returns:
        Unique random filename each time generator is invoked.
    '''
    filenames = set([''])
    result = ''
    while True:
        while result in filenames:
            stem = ''.join(random.choices(string.hexdigits, k=params.filename_length)).lower()
            result = f'{stem}.csv'
        filenames.add(result)
        yield result


def _random_experiment_duration(params: AssayParams, kind: str) -> tuple:
    '''Choose random start date and end date for experiment.

    The start date is uniformly selected from the experiment period.
      The end date is the same as or later than the start date,
      and `None` if the experiment hasn't finished.

    Args:
        params: assay generation parameters.
        kind: experimental procedure used.

    Returns:
        A pair with a start date and either an end date or `None`.
    '''
    start = random.uniform(params.startdate.timestamp(), params.enddate.timestamp())
    start = datetime.fromtimestamp(start, tz=timezone.utc)
    duration = timedelta(days=random.randint(*params.assay_duration))
    end = start + duration
    end = None if end > params.enddate else end
    return start, end


def _random_plates(params: AssayParams, kind: str, sample_id: int, start_id: int, start_date: date, filename_gen: str) -> list:
    '''Generate random plate data.

    Args:
        params: assay generation parameters.
        kind: experimental procedure used.
        sample_id: sample used in the experiment.
        start_id: starting ID of plates.
        start_date: when experiment started.
        filename_gen: random filename generator

    Returns:
        List of dictionaries of plate data.
    '''
    return [
        {
            'plate_id': start_id + i + 1,
            'sample_id': sample_id,
            'date': _random_date_interval(start_date, params.enddate),
            'filename': next(filename_gen),
        }
        for i in range(random.randint(*params.assay_plates))
    ]

def _random_date_interval(start_date: date, end_date: date) -> date:
    '''Choose a random date (inclusive).

    Args:
        start_date: earliest allowed date.
        end_date: last possible date.

    Returns:
        Randomly-selected date.
    '''
    # Convert date to datetime at midnight UTC
    start_datetime = datetime.combine(start_date, time.min).replace(tzinfo=pytz.UTC)
    end_datetime = datetime.combine(end_date, time.max).replace(tzinfo=pytz.UTC)

    choice_timestamp = random.uniform(start_datetime.timestamp(), end_datetime.timestamp())
    choice = datetime.fromtimestamp(choice_timestamp, tz=pytz.UTC)
    return _round_date(choice)

def _round_date(raw: datetime|None) -> date|None:
    '''Round time to whole day.

    Args:
        raw: starting datetime (or `None`).

    Returns:
        Input rounded to nearest whole day (or `None`).
    '''
    return None if raw is None else raw.date()

def _save(outfile: str, result: dict) -> None:
    '''Save or show generated data.

    Args:
        outfile: where to write (or `None` for standard output).
        result: data to write.
    '''
    as_text = json.dumps(result, indent=4, cls=DateTimeEncoder)
    if outfile:
        Path(outfile).write_text(as_text)
    else:
        print(as_text)



'''Generate sample snails with genomes and locations.'''

from argparse import Namespace
import json
import math
import numpy as np
from pathlib import Path
import polars as pl
import random
from geopy.distance import lonlat, distance
import sys

from .params import SampleParams, load_params


LON_LAT_PRECISION = 5
READING_PRECISION = 1
SIZE_PRECISION = 1


def samples(options: Namespace) -> None:
    '''Main driver for snailz snail creation.

    -   options.genomes: genomes data file.
    -   options.grids: grids parameter file.
    -   options.params: path to parameter file (see params.SampleParams for fields).
    -   options.outfile: optional path to saved output file.
    -   options.surveys: survey CSV parameter file.
    -   options.sites: sites CSV parameter file.

    Generated data is written as CSV to the specified output file.

    Args:
        options: see above.
    '''
    assert options.params != options.outfile, 'Cannot use same filename for options and parameters'
    options.params = load_params(SampleParams, options.params)
    options.surveys = pl.read_csv(options.surveys)
    options.sites = pl.read_csv(options.sites)
    random.seed(options.params.seed)

    genomes = json.loads(Path(options.genomes).read_text())
    grids = _load_grids(options)

    samples = _generate_samples(options, genomes, grids)
    _save(options, samples)


def _generate_samples(options: Namespace, genomes: dict, grids: dict) -> pl.DataFrame:
    '''Generate snail samples.

    For each previously-generated genome:

    1.  Select a survey and a random point in that survey's area, and determine if that point is contaminated.
    2.  Determine the range of possible snail sizes based on genotype and contamination.
    3.  Generate a size.
    4.  Append a record to a list that is later converted to a dataframe.

    Args:
        options: see samples().
        genomes: JSON representation of previously-generated genomes.
        grids: key-to-grid dictionary whose grids are NumPy arrays.

    Returns:
        Dataframe with sample ID, survey ID, longitude, latitude, sequence, and snail size.
    '''

    # Generate.
    samples = []
    for i, seq in enumerate(genomes['individuals']):
        survey_id, point, contaminated = _random_geo(options.sites, options.surveys, grids)
        limit = _size_limit(options, genomes, seq, contaminated)
        size = random.uniform(
            options.params.min_snail_size,
            options.params.min_snail_size + options.params.max_snail_size * limit
        )
        samples.append((i + 1, survey_id, point.longitude, point.latitude, seq, size))

    # Convert to dataframe.
    df = pl.DataFrame(samples, schema=('sample_id', 'survey_id', 'lon', 'lat', 'sequence', 'size'), orient='row')
    return df.with_columns(
        lon=df['lon'].round(LON_LAT_PRECISION),
        lat=df['lat'].round(LON_LAT_PRECISION),
        size=df['size'].round(SIZE_PRECISION),
    )


def _load_grids(options: Namespace) -> dict:
    '''Load all grid files.

    Args:
        options: see samples().

    Returns:
        Key-to-NumPy array map of contamination grids.
    '''
    return {
        s: np.loadtxt(Path(options.grids, f'{s}.csv'), dtype=int, delimiter=',')
        for s in set(options.surveys['site_id'])
    }


def _random_geo(sites: pl.DataFrame, surveys: pl.DataFrame, grids: dict) -> tuple:
    '''Select random point from a randomly-selected sample grid.

    1.  Select site.
    2.  Select random grid cell.
    3.  Determine whether that cell is contaminated.
    4.  Use site center point and survey spacing to determine longitude and latitude of cell.

    Args:
        sites: dataframe of site data.
        surveys: dataframe of surveys.
        grids: key-to-grid dictionary whose grids are NumPy arrays.

    Returns:
        Selected survey ID, (lon, lat) point, and whether point is contaminated
    '''
    # Get parameters.
    survey_row = random.randrange(surveys.shape[0])
    survey_id = surveys.item(survey_row, 'survey_id')
    spacing = float(surveys.item(survey_row, 'spacing'))
    site_id = surveys.item(survey_row, 'site_id')
    site_row = sites['site_id'].to_list().index(site_id)

    # Get grid information.
    grid = grids[site_id]
    width, height = grid.shape
    rand_x, rand_y = random.randrange(width), random.randrange(height)
    contaminated = bool(grid[rand_x, rand_x] != 0)

    # Generate point.
    site_lon = sites.item(site_row, 'lon')
    site_lat = sites.item(site_row, 'lat')
    corner = lonlat(site_lon, site_lat)
    rand_x *= spacing
    rand_y *= spacing
    dist = math.sqrt(rand_x**2 + rand_y**2)
    bearing = math.degrees(math.atan2(rand_y, rand_x))
    point = distance(meters=dist).destination(corner, bearing=bearing)

    return survey_id, point, contaminated


def _save(options: Namespace, samples: pl.DataFrame) -> None:
    '''Save results to file or show on standard output.

    Args:
        options: controlling options.
        samples: dataframe of generated samples.
    '''
    if options.outfile:
        samples.write_csv(Path(options.outfile))
    else:
        samples.write_csv(sys.stdout)


def _size_limit(options: Namespace, genomes: dict, seq: str, contaminated: bool) -> float:
    '''Calculate upper bound on snail size.

    If the genome has the significant mutation in the right location
      and the site is contaminated, the snail may have the mutant size.
      Otherwise, it has the normal size.

    Args:
        options: controlling options.
        genomes: JSON containing overall information about genomes.
        seq: specific sequence of this snail.
        contaminated: is sample location contaminated?

    Returns:
        Parameter value for upper bound on normal or mutant snail size.
    '''
    susc_loc = genomes['susceptible_loc']
    susc_base = genomes['susceptible_base']
    if contaminated and (seq[susc_loc] == susc_base):
        return options.params.mutant
    return options.params.normal



'''Mangle readings plate files to simulate poor formatting.'''

from argparse import Namespace
import csv
from pathlib import Path
import random
import sqlite3


# Query used to join records from database.
SELECT = '''
select
    plate.plate_id as plate_id,
    plate.filename as filename,
    plate.date as date,
    staff.staff_id as staff_id,
    staff.personal as personal,
    staff.family as family
from
    plate join performed join staff
on
    (plate.sample_id = performed.sample_id)
    and
    (performed.staff_id = staff.staff_id)
'''


def mangle(options: Namespace) -> None:
    '''Main driver for mangling readings plates.

    -   options.dbfile: path to SQLite database.
    -   options.outdir: output directory.
    -   options.tidy: directory containing tidy readings files.

    Mangled files are written to the specified output directory.
      The files have the same name as the original (tidy) files.

    Args:
        options: see above.
    '''
    con = sqlite3.connect(options.dbfile)
    con.row_factory = sqlite3.Row
    records = list(dict(r) for r in con.execute(SELECT).fetchall())
    random.seed(len(records))
    records = _consolidate(records)
    for rec in records:
        _mangle_file(options, rec)


def _consolidate(records: list) -> list:
    '''Pick a single (plate, staff) pair for each plate.

    Args:
        records: list of (plate, staff) pairs from database.

    Returns:
        One (plate, staff) pair for each plate.
    '''
    grouped = {}
    for r in records:
        if r['plate_id'] not in grouped:
            grouped[r['plate_id']] = []
        grouped[r['plate_id']].append(r)

    result = []
    for group in grouped.values():
        result.append(random.choice(group))
    return result


def _mangle_file(options: Namespace, record: dict) -> None:
    '''Mangle a single file.

    1.  Read file as header and body sections.
    2.  Apply randomly-chosen mangling functions to modify in place.
    3.  Save result.

    Args:
        options: see above.
        record: dictionary of database query results for a single record.
    '''
    sections = _read_sections(options, record['filename'])
    for func in (_do_staff_name, _do_date, _do_footer, _do_indent,):
        if random.random() < func.prob:
            func(record, sections)
    _write_sections(options, record['filename'], sections)


def _do_date(record: dict, sections: dict) -> None:
    '''Mangle data in place by adding date in header.

    Args:
        record: entire record data.
        sections: dictionary of header, body, and footer.
    '''
    row = [''] * len(sections['header'][0])
    row[0] = 'Date'
    row[1] = record['date']
    sections['header'].append(row)
_do_date.prob = 0.1


def _do_footer(record: dict, sections: dict) -> None:
    '''Mangle data in place by adding a footer.

    Args:
        record: entire record data.
        sections: dictionary of header, body, and footer.
    '''
    blank = [''] * len(sections['header'][0])
    foot = [''] * len(sections['header'][0])
    foot[0] = record['staff_id']
    sections['footer'] = [blank, foot]
_do_footer.prob = 0.1


def _do_indent(record: dict, sections: dict) -> None:
    '''Mangle data in place by indenting all rows by one space

    Args:
        record: entire record data.
        sections: dictionary of header, body, and footer.
    '''
    for section in sections.values():
        for row in section:
            row.insert(0, '')
_do_indent.prob = 0.1


def _do_staff_name(record: dict, sections: dict) -> None:
    '''Mangle data in place by adding staff name.

    Args:
        record: entire record data.
        sections: dictionary of header, body, and footer.
    '''
    sections['header'][0][-2] = f'{record["personal"]} {record["family"]}'
_do_staff_name.prob = 0.1


def _read_sections(options: Namespace, filename: str) -> dict:
    '''Read tidy readings file and split into sections.

    Args:
        options: see above.
        filename: file to read from.

    Returns:
        Dictionary with header, head-to-body spacing, body, and footer (empty in tidy file).
    '''
    with open(Path(options.tidy, filename), 'r') as raw:
        rows = [row for row in csv.reader(raw)]
    return {
        'header': rows[0:1],
        'headspace': rows[1:2],
        'body': rows[2:],
        'footer': []
    }


def _write_sections(options: Namespace, filename: str, sections: dict) -> None:
    '''Write sections of mangled file to file.

    Args:
        options: see above.
        filename: file to write to.
        sections: dictionary of header, head-to-body spacing, body, and footer.
    '''
    with open(Path(options.outdir, filename), 'w') as raw:
        writer = csv.writer(raw, lineterminator='\n')
        for section in sections.values():
            writer.writerows(section)



'''Generate snailz SQLite database from data files.'''

from argparse import Namespace
import json
import polars as pl
import sqlite3

from .params import DATE_FORMAT


# Database schema.
SCHEMA = '''\
DROP TABLE IF EXISTS site;
CREATE TABLE site (
	site_id TEXT PRIMARY KEY,
	lon FLOAT NOT NULL,
	lat FLOAT NOT NULL
);
DROP TABLE IF EXISTS survey;
CREATE TABLE survey (
	survey_id INTEGER PRIMARY KEY,
	site_id TEXT NOT NULL,
	date DATE,
	FOREIGN KEY(site_id) REFERENCES site(site_id)
);
DROP TABLE IF EXISTS sample;
CREATE TABLE sample (
	sample_id INTEGER PRIMARY KEY,
	survey_id INTEGER NOT NULL,
	lon FLOAT NOT NULL,
	lat FLOAT NOT NULL,
	sequence TEXT NOT NULL,
	size FLOAT NOT NULL,
	FOREIGN KEY(survey_id) REFERENCES survey(survey_id)
);
DROP TABLE IF EXISTS experiment;
CREATE TABLE experiment (
	sample_id INTEGER PRIMARY KEY,
	kind TEXT NOT NULL,
	start DATE NOT NULL,
	"end" DATE,
	FOREIGN KEY(sample_id) REFERENCES sample(sample_id)
);
DROP TABLE IF EXISTS staff;
CREATE TABLE staff (
	staff_id INTEGER PRIMARY KEY,
	personal TEXT NOT NULL,
	family TEXT NOT NULL
);
DROP TABLE IF EXISTS performed;
CREATE TABLE performed (
	staff_id INTEGER NOT NULL,
	sample_id INTEGER NOT NULL,
	PRIMARY KEY(staff_id, sample_id)
);
DROP TABLE IF EXISTS plate;
CREATE TABLE plate (
	plate_id INTEGER PRIMARY KEY,
	sample_id INTEGER NOT NULL,
	date DATE NOT NULL,
	filename TEXT NOT NULL,
	FOREIGN KEY(sample_id) REFERENCES sample(sample_id)
);
DROP TABLE IF EXISTS invalidated;
CREATE TABLE invalidated (
	plate_id INTEGER NOT NULL,
	staff_id INTEGER NOT NULL,
	date DATE,
	FOREIGN KEY(plate_id) REFERENCES plate(plate_id),
	FOREIGN KEY(staff_id) REFERENCES staff(staff_id),
	UNIQUE(plate_id, staff_id)
);
'''


def db(options: Namespace) -> None:
    '''Main driver for database generator.

    Args:
        options: controlling options.

    -   dbfile: output database file.
    -   assays: generated assay data.
    -   samples: generated sample data.
    -   sites: site parameters.
    -   staff: generated staff data.
    -   surveys: survey parameters.
    '''
    _create_schema(options.dbfile)

    url = f'sqlite:///{options.dbfile}'

    _csv_to_db(url, 'staff', options.staff)
    _csv_to_db(url, 'sample', options.samples)
    _csv_to_db(url, 'site', options.sites)
    _csv_to_db(url, 'survey', options.surveys, dates=['date'], columns=['survey_id', 'site_id', 'date'])

    assays = json.load(open(options.assays, 'r'))
    _json_to_db(url, assays, 'experiment', dates=['start', 'end'])
    _json_to_db(url, assays, 'performed')
    _json_to_db(url, assays, 'plate', dates=['date'])
    _json_to_db(url, assays, 'invalidated', dates=['date'])


def _create_schema(dbfile: str) -> None:
    '''Create database schema.

    Args:
        dbfile: database file path.
    '''
    connection = sqlite3.connect(dbfile)
    connection.executescript(SCHEMA)


def _csv_to_db(url: str, name: str, source: str, dates: list = [], columns: list = []) -> None:
    '''Fill table from CSV.

    Args:
        url: connection URL for database.
        name: table name.
        source: path to CSV source file.
        dates: columns to convert to dates.
        columns: subset of columns to store in table.
    '''
    df = pl.read_csv(source)
    if columns:
        df = df[columns]
    df = _convert_dates(df, dates)
    df.write_database(name, url, if_table_exists='append')


def _json_to_db(url: str, data: dict, name: str, dates: list = []) -> None:
    '''Fill table from JSON.

    Args:
        url: connection URL for database.
        data: JSON data.
        name: table name (must be field in `data`).
        dates: columns to convert to dates.
    '''
    df = pl.DataFrame(data[name])
    df = _convert_dates(df, dates)
    df.write_database(name, url, if_table_exists='append')


def _convert_dates(df: pl.DataFrame, dates: list) -> pl.DataFrame:
    '''Convert text columns to dates.

    Args:
        df: input dataframe.
        dates: list of names of columns to convert.

    Returns:
        Updated dataframe.
    '''
    for colname in dates:
        df = df.with_columns(pl.col(colname).str.to_date(DATE_FORMAT))
    return df



r'''Snailz is a collection of synthetic data generators for use in teaching.

To re-create all of the data, run the following commands:

# Create copies of default parameter files (replacing './params' as necessary).
mkdir params
snailz params \
  --outdir params

# Make required output directories (replacing '.' as necessary).
mkdir -p data/grids data/designs data/readings

# Synthesize genomes for snails from survey sites.
snailz genomes \
  --outfile data/genomes.json \
  --params params/genomes.json

# Synthesize grids showing pollution at survey sites.
snailz grid \
  --grids params/grids.json \
  --outdir data/grids \
  --sites params/sites.csv

# Synthesize descriptions of snails collected in surveys.
snailz samples \
  --genomes data/genomes.json \
  --grids data/grids \
  --outfile data/samples.csv \
  --params params/samples.json \
  --sites params/sites.csv \
  --surveys params/surveys.csv

# Generate map of survey locations.
snailz survey \
  --outfile data/survey.png \
  --samples data/samples.csv

# Synthesize assay experiment descriptions.
snailz assays \
  --genomes data/genomes.json \
  --outfile data/assays.json \
  --params params/assays.json \
  --samples data/samples.csv

# Store information generated so far in a SQLite database.
snailz db \
  --dbfile data/lab.db \
  --assays data/assays.json \
  --samples data/samples.csv \
  --sites params/sites.csv \
  --surveys params/surveys.csv

# Synthesize assay plates used in those experiments.
snailz plates \
  --assays data/assays.json \
  --designs data/designs \
  --params params/assays.json \
  --readings data/readings

# Synthesize 'raw' reading files.
snailz mangle \
  --dbfile data/lab.db \
  --tidy data/readings \
  --outdir data/mangled
'''

from .assays import assays
from .db import db
from .genomes import genomes
from .grid import grid
from .mangle import mangle
from .plates import plates
from .samples import samples
from .surveymap import surveymap

__all__ = [
    'assays',
    'db',
    'genomes',
    'grid',
    'mangle',
    'plates',
    'samples',
    'surveymap',
]
